#!/usr/bin/env python3

import os
import unittest
from shlex import quote

from bashmocker import BashMocker

class T(unittest.TestCase):
    """Internal testing for BashMocker helper
    """
    shell = None

    def test_bin_mocker(self):
        with BashMocker('foo', 'bad', shell=self.shell) as bm:
            bm.add_mock('bad', fail=True)

            res1 = bm.runscript('true')
            self.assertEqual(res1, 0)
            self.assertEqual(bm.last_stdout, '')
            self.assertEqual(bm.last_stderr, '')
            self.assertEqual(bm.last_calls, bm.empty_calls())
            self.assertEqual(bm.last_calls, dict(foo=[], bad=[]))

            # Now something that actually calls things
            res2 = bm.runscript('echo 123 ; foo foo args ; echo 456 ; ( echo 888 >&2 ) ; bad dog ; bad doggy')
            self.assertEqual(res2, 1)
            self.assertEqual(bm.last_stdout.rstrip().split('\n'), ['123', '456'])
            self.assertEqual(bm.last_stderr.rstrip().split('\n'), ['888'])
            self.assertEqual(bm.last_calls['foo'], [ ["foo", "args"] ])
            self.assertEqual(bm.last_calls['bad'], [ ["dog"], ["doggy"] ])

            # Test that everything resets properly
            res3 = bm.runscript('true')
            self.assertEqual(res3, 0)
            self.assertEqual(bm.last_stdout, '')
            self.assertEqual(bm.last_stderr, '')
            self.assertEqual(bm.last_calls, dict(foo=[], bad=[]))

    def test_cmd_as_list(self):
        """When bm.runscript() is called with a list arg it should bypass the shell.
        """
        with BashMocker('foo', 'bad', shell=self.shell) as bm:
            bm.add_mock('bad', fail=True)

            # Same as above
            res1 = bm.runscript(['true'])
            self.assertEqual(res1, 0)
            self.assertEqual(bm.last_stdout, '')
            self.assertEqual(bm.last_stderr, '')
            self.assertEqual(bm.last_calls, bm.empty_calls())
            self.assertEqual(bm.last_calls, dict(foo=[], bad=[]))

            # Now echo some stuff.
            res2 = bm.runscript(['echo', '123', '\n', '* <!!!"'])
            self.assertEqual(res2, 0)
            self.assertEqual(bm.last_stderr, '')
            self.assertEqual(bm.last_stdout, '123 \n * <!!!"\n')

            # And run the mocked scripts
            res3 = bm.runscript(['foo'])
            self.assertEqual(res3, 0)
            self.assertEqual(bm.last_calls['foo'], [ [] ])

            res4 = bm.runscript(['/usr/bin/env', 'bad', 'dog', ')))))'])
            self.assertEqual(res4, 1)
            self.assertEqual(bm.last_calls['foo'], [])
            self.assertEqual(bm.last_calls['bad'], [ ["dog", ")))))"] ])

    def test_character_escaping(self):
        with BashMocker('foo', shell=self.shell) as bm:

            # Some exotic looking args
            args = [ [ "a", "b", "c", "a b c" ],
                     [ ''' ""''"<>!!! ''' ],
                     [ '\n\n', '', '\t\t \n\r\r' ],
                     [ '-' ] ]

            script = ' ; '.join([ ' '.join(['foo'] + [quote(a) for a in alist])
                                  for alist in args ])

            bm.runscript(script)

            self.assertEqual(bm.last_stdout, '')
            self.assertEqual(bm.last_stderr, '')
            self.assertEqual(bm.last_calls, dict(foo=args))

    def test_side_effect(self):
        """We can add a side_effect to our mock.
        """
        bm = BashMocker(shell=self.shell)
        self.addCleanup(bm.cleanup)

        # Side effect should happen but should not affect the return value.
        bm.add_mock('this', side_effect="echo THIS ; false")
        bm.add_mock('that', side_effect="echo THAT >&2 ; true", fail=True)

        # Unless the side effect explicitly calls 'exit'.
        bm.add_mock('theother', side_effect="echo THEOTHER ; exit $1")

        res1 = bm.runscript('this')
        self.assertEqual(res1, 0)
        self.assertEqual(bm.last_stdout, 'THIS\n')

        res2 = bm.runscript('that')
        self.assertEqual(res2, 1)
        self.assertEqual(bm.last_stderr, 'THAT\n')

        res3 = bm.runscript('theother 42')
        self.assertEqual(res3, 42)
        self.assertEqual(bm.last_stdout, 'THEOTHER\n')
        self.assertEqual(bm.last_calls, dict( this = [],
                                              that = [],
                                              theother = [[ '42' ]] ))

    def test_mock_abs_path(self):
        """It's possible, if hacky, to mock out even commands referred to by
           full path. Note that this will only work for things called from
           BASH, not for things called indirectly like "env /bin/foo". And
           it won't work on DASH or with BASH in compatibility mode as these don't
           read BASH_ENV.
        """
        bm = BashMocker(shell=self.shell)
        self.addCleanup(bm.cleanup)

        bm.add_mock('/bin/false', side_effect="echo THIS")

        res1 = bm.runscript('/bin/false 123')
        self.assertEqual(res1, 0)
        self.assertEqual(bm.last_calls['/bin/false'], [ ['123'] ])
        self.assertEqual(bm.last_stdout, 'THIS\n')

        #Should also work if called from a sub-script.
        bm.add_mock('woo',  side_effect="/bin/false 789")
        bm.add_mock('/bin/wibble',  side_effect="woo 456")

        res2 = bm.runscript('/bin/wibble 123')
        self.assertEqual(res2, 0)
        self.assertEqual(bm.last_calls['/bin/wibble'], [ ['123'] ])
        self.assertEqual(bm.last_calls['woo'], [ ['456'] ])
        self.assertEqual(bm.last_calls['/bin/false'], [ ['789'] ])
        self.assertEqual(bm.last_stdout, 'THIS\n')

        #Referring to a command stored in a var is OK
        res3 = bm.runscript('cmd=/bin/false ; "$cmd" 123')
        self.assertEqual(res3, 0)
        self.assertEqual(bm.last_stdout, 'THIS\n')

        #But calling a command via 'env' does call the actual command
        res4 = bm.runscript('env /bin/false 123')
        self.assertEqual(res4, 1)
        self.assertEqual(bm.last_stdout, '')

    def test_mock_abs_2(self):
        """I was worried about a case where calling a function with no
           args leaves the script args in "$@" but this doesn't seem to
           happen. Good.
        """
        bm = BashMocker(shell=self.shell)
        self.addCleanup(bm.cleanup)

        bm.add_mock('bibble',  side_effect="/no/such/thing")
        bm.add_mock("/no/such/thing", side_effect='echo "$@"')

        res = bm.runscript('bibble abc 123')
        self.assertEqual(bm.last_calls['bibble'], [ ['abc', '123'] ])
        self.assertEqual(bm.last_calls['/no/such/thing'], [ [] ])
        self.assertEqual(bm.last_stdout, '\n')

    def test_fail_mock(self):
        """I was seeing weird behaviour on Ubuntu. Actually I don't think
           binmocker was to blame but I'm keeping this test anyway.
        """
        with BashMocker(shell=self.shell) as bm:
            bm.add_mock('foo')
            res = bm.runscript('foo')
            self.assertEqual(res, 0)
            self.assertEqual(bm.last_stdout, '')

            res = bm.runscript('foo ; echo $?')
            self.assertEqual(res, 0)
            self.assertEqual(bm.last_stdout, '0\n')

            # Now make it fail...
            bm.add_mock('foo', fail=True)

            res = bm.runscript('foo')
            self.assertEqual(res, 1)
            self.assertEqual(bm.last_stdout, '')

            res = bm.runscript('foo ; echo $?')
            self.assertEqual(res, 0)
            self.assertEqual(bm.last_stdout, '1\n')

    def test_unlogged(self):
        """We can add a mock that is never logged.
        """
        with BashMocker('foo', shell=self.shell) as bm:
            bm.add_mock('unlogged', log=False)

            res1 = bm.runscript('foo 1234 ; unlogged 5678')
            self.assertEqual(res1, 0)
            self.assertEqual(bm.last_stdout, '')
            self.assertEqual(bm.last_stderr, '')
            self.assertEqual(bm.empty_calls(), dict(foo=[]))
            self.assertEqual(bm.last_calls, dict(foo=[['1234']]))

    def test_unlogged_func(self):
        """Unlogged functions should be OK too
        """
        with BashMocker('foo', shell=self.shell) as bm:
            bm.add_mock('unlogged', log=False)
            bm.add_mock('bar/bar', log=True)
            bm.add_mock('unlogged/unlogged', log=False)

            res2 = bm.runscript('bar/bar 1234 ; unlogged/unlogged 5678')
            self.assertEqual(bm.empty_calls(), {'foo':[], 'bar/bar':[]})
            self.assertEqual(bm.last_calls, {'foo':[], 'bar/bar':[['1234']]})

# On Debian-type systems SH will normally be DASH. On systems where SH is BASH then
# BASH will behave differently depending how it's called. In either case the mock-by-function
# hack shouldn't work.
class T_sh(T):
    """Test with explicit calls to /bin/sh
    """
    shell = '/bin/sh'

    # This hack only works in BASH. Maybe there is another hack for DASH?
    @unittest.expectedFailure
    def test_mock_abs_path(self):
        super().test_mock_abs_path()

    @unittest.expectedFailure
    def test_mock_abs_2(self):
        super().test_mock_abs_2()

    @unittest.expectedFailure
    def test_unlogged_func(self):
        super().test_unlogged_func()

@unittest.skipUnless(os.path.exists("/bin/dash") , "no /bin/dash")
class T_dash(T):
    """Test with explicit calls to DASH
    """
    shell = '/bin/dash'

    # This hack only works in BASH. Maybe there is another hack for DASH?
    @unittest.expectedFailure
    def test_mock_abs_path(self):
        super().test_mock_abs_path()

    @unittest.expectedFailure
    def test_mock_abs_2(self):
        super().test_mock_abs_2()

    @unittest.expectedFailure
    def test_unlogged_func(self):
        super().test_unlogged_func()

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

"""bashmocker - a hack for intercepting subprocess calls to test shell scripts in Python"""

import os
import subprocess
from tempfile import mkdtemp
from shutil import rmtree


class BashMocker:
    """A helper class that provides a way to replace tools with dummy
       calls and then summarize what was called.
       The idea is to make something vaguely similar to Test::Mock. I
       wouldn't say this is fully robust but for unit tests it does the job.

       Given a shell script named foo.sh:
         #!/bin/bash
         mycmd 1
         mycmd 2

       We can run the script with commands mocked out:
         with BashMocker('mycmd') as bm:
            bm.runscript('foo.sh')

            #Check that foo.sh called mycmd twice.
            assert len(bm.last_calls['mycmd']) == 2

       I made this for testing BASH scripts but in fact any program that calls
       other programs and respects env['PATH'] is testable this way.
    """

    def __init__(self, *mocks, shell=None):
        self.mock_bin_dir = mkdtemp()

        # The shell may be changed to /bin/sh or /bin/dash or but the normal default of
        # /bin/bash should be fine regardless of what you are testing.
        self._shell = shell or "/bin/bash"
        self._sh_env_val = None

        self._make_mockscript("_MOCK", 0)
        self._make_mockscript("_MOCKF", 1)
        self._make_mockscript("_MOCK_NOLOG", 0, log=False)
        self._make_mockscript("_MOCKF_NOLOG", 1, log=False)

        self.mocks = set()
        self.mocks_nolog = set()
        for m in mocks:
            self.add_mock(m)

        self.last_calls = None
        self.last_stderr = None
        self.last_stdout = None

    def _make_mockscript(self, mockname, retcode, log=True, side_effect="#NOP"):
        """Internal function for making mock scripts.
        """

        # To make the saving of args robust I had this idea:
        # for x in "$@" ; do _fs+=(%q) ; done
        # printf "%q %d ${_fs[*]}\n" "$0" "$#" "$@"
        # But using zeros seems better.
        # Also I want this to work in both BASH and DASH which is tricky.
        shell = self._shell
        if log:
            mockscript = r'''
             #!{shell}
             _fs=%s\\0%d\\0 ; for x in "$@" ; do _fs="$_fs"%s\\0 ; done ; _fs="$_fs"\\n
             printf "$_fs" "$(basename "$0")" "$#" "$@" >> "$(dirname "$0")"/_MOCKCALLS
             {side_effect}
             exit {retcode}
            '''
        else:
            mockscript = r'''
             #!{shell}
             {side_effect}
             exit {retcode}
            '''
        mockscript = mockscript.format(**locals())

        with open(os.path.join(self.mock_bin_dir, mockname), 'w') as fh:
            print(mockscript.strip(), file=fh)

            # copy R bits to X to achieve chmod +x
            mode = os.stat(fh.fileno()).st_mode
            os.chmod(fh.fileno(), mode | (mode & 0o444) >> 2)

    def _add_mockfunc(self, funcname, retcode, log=True, side_effect="#NOP"):
        """Internal function for writing mock functions.
           Note these will only apply to scripts that explicitly use #!/bin/bash
           as the interpreter, not #!/bin/sh or #!/bin/dash or anything
           called indirectly within the script, like 'env /bin/true ...'.
        """
        # This hack only works on BASH by setting BASH_ENV. DASH does not read any preable for
        # non-interactive scripts so there seems to be no easy way to poke one in.
        if not self._shell.endswith('/bash'):
            raise RuntimeError("Only BASH allows setting a preamble for scripts")

        if log:
            mockfunc = r'''
             {funcname}(){{
             local _fs
             _fs=%s\\0%d\\0 ; for x in "$@" ; do _fs="$_fs"%s\\0 ; done ; _fs="$_fs"\\n
             printf "$_fs" '{funcname}' "$#" "$@" >> "$(dirname "$BASH_SOURCE")"/_MOCKCALLS
             {side_effect}
             return {retcode} ; }}
            '''
        else:
            mockfunc = r'''
             {funcname}(){{
             {side_effect}
             return {retcode} ; }}
            '''
        mockfunc = mockfunc.format(**locals())

        with open(os.path.join(self.mock_bin_dir, '_BASH_ENV'), 'a') as fh:
            print(mockfunc.strip(), file=fh)

        self._sh_env_val = os.path.join(self.mock_bin_dir, "_BASH_ENV")

    def add_mock(self, mock, fail=False, log=True, side_effect=None):
        """Make a symlink named <mock> so the mock script will get called in
           place of the given command.
           Except if mock contains a / character - then bashmocker will try to use
           functions instead.

           If log is False then calls to this command will not be logged.

           If a side_effect is specified this shell code will be run as part of the
           script.
        """
        if '/' in mock:
            # We can still mock these by defining BASH functions.
            # TODO - should we do this for builtins too?
            self._add_mockfunc(mock, 1 if fail else 0, log=log, side_effect=side_effect)

        else:

            symlink = os.path.join(self.mock_bin_dir, mock)

            if side_effect:
                # We need a special script then
                target = "_MOCK_" + mock
                self._make_mockscript(target, 1 if fail else 0, log=log, side_effect=side_effect)
            else:
                if log:
                    target = "_MOCKF" if fail else "_MOCK"
                else:
                    target = "_MOCKF_NOLOG" if fail else "_MOCK_NOLOG"

            # If the link already exists, remove it
            try:
                os.unlink(symlink)
            except FileNotFoundError:
                pass

            os.symlink(target, symlink)

        if log:
            self.mocks.add(mock)
        else:
            self.mocks_nolog.add(mock)

    def cleanup(self):
        """Clean up
        """
        rmtree(self.mock_bin_dir)
        self.mock_bin_dir = None

    def runscript(self, cmd, set_path=True, env=None):
        """Runs the specified command, which may contain shell syntax if
           it is a string or else may be a list of literal [cmd, arg1, arg2, ...]
           and captures the output and the commands that were invoked.

           By default, the mock scripts will be prepended to the PATH, but you
           can alternatively specify set_path=False in which case you take
           responsibility for adding bm.mock_bin_dir to the PATH.

           If env is supplied, it must be a dict of strings. Items in env will
           be added to the execution environment.
        """
        # Cleanup _MOCKCALLS if found
        calls_file = os.path.join(self.mock_bin_dir, "_MOCKCALLS")
        try:
            os.unlink(calls_file)
        except FileNotFoundError:
            pass

        full_env = None
        if env:
            full_env = os.environ.copy()
            full_env.update(env)

        if set_path:
            full_env = full_env or os.environ.copy()
            if full_env.get('PATH'):
                full_env['PATH'] = os.path.abspath(self.mock_bin_dir) + ':' + full_env['PATH']
            else:
                full_env['PATH'] = os.path.abspath(self.mock_bin_dir)

            if self._sh_env_val:
                # Note - interactive scripts shouldn't normally depend on BASH_ENV,
                # so complain if I'm clobbering it. Ditto for ENV, but we don't need that
                # as BASH in compatibility mode won't set the functions we need.
                if full_env.get('BASH_ENV'):
                    raise RuntimeError("BASH_ENV was already set")

                # Rather than working out which is needed, set both.
                full_env['BASH_ENV'] = self._sh_env_val

        use_shell = (type(cmd) == str)
        p = subprocess.Popen(cmd,
                             shell = use_shell,
                             stdout = subprocess.PIPE,
                             stderr = subprocess.PIPE,
                             universal_newlines = True,
                             env = full_env,
                             executable = self._shell if use_shell else None,
                             close_fds = True)

        self.last_stdout, self.last_stderr = p.communicate()

        # Fish the MOCK calls out of _MOCKCALLS
        # Each line is \0 delimited but there may also be embedded newlines in
        # the arguments, so we read it in a funny way. I could just slurp the file
        # instead, I guess.
        calls = self.empty_calls()
        try:
            with open(calls_file, newline='\n') as fh:
                for aline in fh:
                    mock_name, mock_argc, mock_argv = aline.split('\0', 2)
                    while mock_argv.count('\0') < int(mock_argc):
                        # Must be an embedded newline; pull the next line
                        mock_argv += next(fh)
                    # The line should end in \0 so discard the last ''
                    mock_args = mock_argv.rstrip('\n').split('\0')[:-1]
                    calls[mock_name].append(mock_args)
        except FileNotFoundError:
            # So, nothing ran
            pass

        self.last_calls = calls

        # Return whatever the process returned
        return p.returncode

    def empty_calls(self):
        """Get the baseline dict of calls. Useful for tests to assert that
             bm.last_calls == bm.empty_calls()
           ie. nothing happened.
        """
        return { m : [] for m in self.mocks }

    # Allow the class to be used in a "with" construct, though
    # in a TestCase you probably want to use setup/teardown.
    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.cleanup()
